"""
Submit handler for the Fastlane project.

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""
__VERSION__ = "1.0"
__DATE__="01/May/2023"

import json
from dataclasses import dataclass, asdict
from typing import List, Any, Dict

import requests
from brownie.network.transaction import TransactionReceipt
from eth_utils import to_hex
from web3._utils.threads import Timeout
from web3._utils.transactions import fill_nonce
from web3.contract import ContractFunction
from web3.exceptions import TimeExhausted
from web3.types import TxParams

#import fastlane_bot.config as c
from .routehandler import RouteStruct
from ..data.abi import ERC20_ABI
from fastlane_bot.config import Config

@dataclass
class TxSubmitHandlerBase:
    __VERSION__=__VERSION__
    __DATE__=__DATE__
    
@dataclass
class TxSubmitHandler(TxSubmitHandlerBase):
    """
    A class that handles the submission of transactions to the blockchain.

    Attributes
    ----------
    route_struct: List[str]
        The route structure for a transaction. As required by the `BancorArbitrage` contract.
    src_amount: int
        The source amount for a transaction. (in wei)
    src_address: str
        The source address for a transaction. (checksummed)

    Methods
    -------
    _get_deadline(self) -> int:
        Gets the deadline for a transaction.
    _get_transaction(self, tx_details: TxParams) -> TxParams:
        Gets the transaction details for a given transaction.
    _get_transaction_receipt(self, tx_hash: str, timeout: int = DEFAULT_TIMEOUT) -> TransactionReceipt:
        Gets the transaction receipt for a given transaction.
    _get_transaction_receipt_with_timeout(self, tx_hash: str, timeout: int = DEFAULT_TIMEOUT) -> TransactionReceipt:
        Gets the transaction receipt for a given transaction with a timeout.

    """
    __VERSION__=__VERSION__
    __DATE__=__DATE__
    
    ConfigObj: Config
    route_struct: List[RouteStruct] = None
    src_address: str = None
    src_amount: int = None




    def __post_init__(self):
        self.w3 = self.ConfigObj.w3
        self.arb_contract = self.ConfigObj.BANCOR_ARBITRAGE_CONTRACT
        self.bancor_network_info = self.ConfigObj.BANCOR_NETWORK_INFO_CONTRACT
        # self.token_contract = Contract.from_abi(
        #     name="Token",
        #     address=self.src_address,
        #     abi=ERC20_ABI,
        # )
        self.token_contract = self.ConfigObj.w3.eth.contract(address=self.src_address, abi=ERC20_ABI)

    def _get_deadline(self) -> int:
        """
        Gets the deadline for a transaction.

        Returns
        -------
        int
            The deadline.
        """
        return (
            self.w3.eth.getBlock(self.w3.eth.block_number).timestamp
            + self.ConfigObj.DEFAULT_BLOCKTIME_DEVIATION
        )

    def _get_transaction(self, tx_details: TxParams) -> TxParams:
        """
        Gets the transaction details for a given transaction.

        Parameters
        ----------
        tx_details: TxParams
            The transaction details.

        Returns
        -------
        TxParams
            The transaction details.
        """
        return fill_nonce(self.ConfigObj.w3, tx_details)

    def _get_transaction_receipt(
        self, tx_hash: str, timeout: int
    ) -> TransactionReceipt:
        """
        Gets the transaction receipt for a given transaction hash.

        Parameters
        ----------
        tx_hash: str
            The transaction hash.
        timeout: int
            The timeout for the transaction receipt.

        Returns
        -------
        TransactionReceipt
            The transaction receipt.
        """
        return self.w3.eth.wait_for_transaction_receipt(tx_hash, timeout=timeout)

    def _get_transaction_receipt_with_timeout(
        self, tx_hash: str, timeout: int, poll_latency: int = 0.1
    ) -> TransactionReceipt:
        """
        Gets the transaction receipt for a given transaction hash.

        Parameters
        ----------
        tx_hash: str
            The transaction hash.
        timeout: int
            The timeout for the transaction receipt.
        poll_latency: int
            The poll latency for the transaction receipt.

        Returns
        -------
        TransactionReceipt
            The transaction receipt.
        """
        with Timeout(timeout) as _timeout:
            while True:
                tx_receipt = self.ConfigObj.w3.eth.getTransactionReceipt(tx_hash)
                if tx_receipt is not None:
                    return tx_receipt
                _timeout.sleep(poll_latency)

    def _get_transaction_receipt_with_timeout_and_retries(
        self,
        tx_hash: str,
        timeout: int,
        poll_latency: int = 0.1,
        retries: int = 5,
    ) -> TransactionReceipt:
        """
        Gets the transaction receipt for a given transaction hash.

        Parameters
        ----------
        tx_hash: str
            The transaction hash.
        timeout: int
            The timeout for the transaction receipt.
        poll_latency: int
            The poll latency for the transaction receipt.
        retries: int
            The number of retries for the transaction receipt.

        Returns
        -------
        TransactionReceipt
            The transaction receipt.
        """
        for _ in range(retries):
            try:
                return self._get_transaction_receipt_with_timeout(
                    tx_hash, timeout, poll_latency
                )
            except TimeExhausted:
                continue
        raise TimeExhausted(f"Transaction {tx_hash} not found after {retries} retries")

    def _get_gas_price(self) -> int:
        """
        Gets the gas price for the transaction.

        Returns
        -------
        int
            The gas price for the transaction.
        """
        return int(self.ConfigObj.w3.eth.gas_price * self.ConfigObj.DEFAULT_GAS_PRICE_OFFSET)

    def _get_gas(self, tx_function: ContractFunction, address: str) -> int:
        """
        Gets the gas for the transaction.

        Parameters
        ----------
        tx_function: ContractFunction
            The transaction function.

        Returns
        -------
        int
            The gas for the transaction.
        """
        return tx_function.estimateGas({"from": address}) + 10000

    @property
    def headers(self):
        return {"accept": "application/json", "content-type": "application/json"}

    @staticmethod
    def _get_payload(method: str, params: [] = None) -> Dict:
        if method in {"eth_estimateGas", "eth_sendPrivateTransaction"}:
            return {"id": 1, "jsonrpc": "2.0", "method": method, "params": params}
        else:
            return {"id": 1, "jsonrpc": "2.0", "method": method}

    def _get_max_priority_fee_per_gas_alchemy(self):
        return self._query_alchemy_api_gas_methods(method="eth_maxPriorityFeePerGas")

    def _get_eth_gas_price_alchemy(self):
        return self._query_alchemy_api_gas_methods(method="eth_gasPrice")

    def _get_gas_estimate_alchemy(self, params: []):
        return self._query_alchemy_api_gas_methods(
            method="eth_estimateGas", params=params
        )

    def _query_alchemy_api_gas_methods(self, method: str, params: list = None):
        response = requests.post(
            self.ConfigObj.ALCHEMY_API_URL,
            json=self._get_payload(method=method, params=params),
            headers=self.headers,
        )
        return int(json.loads(response.text)["result"].split("0x")[1], 16)

    def _get_tx_details(self):
        """
        Gets the transaction details for the transaction. (testing purposes)
        """
        return {
            "gasPrice": self.ConfigObj.DEFAULT_GAS_PRICE,
            "gas": self.ConfigObj.DEFAULT_GAS,
            "from": self.w3.toChecksumAddress(self.ConfigObj.FASTLANE_CONTRACT_ADDRESS),
            "nonce": self.w3.eth.get_transaction_count(
                self.w3.toChecksumAddress(self.ConfigObj.FASTLANE_CONTRACT_ADDRESS)
            ),
        }

    def submit_transaction_tenderly(
        self, route_struct: List[RouteStruct], src_address: str, src_amount: int
    ) -> Any:
        """
        Submits a transaction to the network.

        Parameters
        ----------
        tx_details: TxParams
            The transaction details.
        key: str
            The private key.

        Returns
        -------
        str
            The transaction hash.
        """
        # route_struct = [asdict(r) for r in route_struct]
        #for r in route_struct:
            #print(r)
        #     print("\n")
        # print(
        #     f"Submitting transaction to Tenderly...src_amount={src_amount} src_address={src_address}"
        #)
        address = self.ConfigObj.w3.toChecksumAddress(self.ConfigObj.BINANCE14_WALLET_ADDRESS)
        return self.arb_contract.functions.flashloanAndArb(
            route_struct, src_address, src_amount
        ).transact(
            {
                "gas": self.ConfigObj.DEFAULT_GAS,
                "from": address,
                "nonce": self._get_nonce(address),
                "gasPrice": 0,
            }
        )

    def _get_transaction_details(
        self,
        tx_function: ContractFunction,
        tx_params: TxParams,
        from_address: str,
        to_address: str,
    ) -> TxParams:
        """
        Gets the transaction details for the transaction.

        Parameters
        ----------
        tx_function: ContractFunction
            The transaction function.
        tx_params: TxParams
            The transaction parameters.

        Returns
        -------
        TxParams
            The transaction details.
        """
        return {
            "from": from_address,
            "to": to_address,
            "gas": self._get_gas(tx_function),
            "gasPrice": self._get_gas_price(),
            "value": tx_params.get("value", 0),
            "nonce": tx_params.get("nonce", None),
            "chainId": self.w3.eth.chain_id,
        }

    def _get_transaction_hash(self, tx_details: TxParams, key: str) -> str:
        """
        Gets the transaction hash for the transaction.

        Parameters
        ----------
        tx_details: TxParams
            The transaction details.
        key: str
            The private key.

        Returns
        -------
        str
            The transaction hash.
        """
        return self.ConfigObj.w3.eth.send_raw_transaction(
            to_hex(self.ConfigObj.w3.eth.account.sign_transaction(tx_details, key).rawTransaction)
        )

    def _get_nonce(self, address: str) -> int:
        """
        Gets the nonce for the transaction.

        Parameters
        ----------
        address: str
            The address.

        Returns
        -------
        int
            The nonce for the transaction.
        """
        return self.ConfigObj.w3.eth.getTransactionCount(address)

    _get_gas_estimate = _get_gas

    # build transaction
    def _build_transaction(
        self,
        tx_function: ContractFunction,
        tx_params: TxParams,
        from_address: str,
        to_address: str,
    ) -> TxParams:
        """
        Builds the transaction details for the transaction.

        Parameters
        ----------
        tx_function: ContractFunction
            The transaction function.
        tx_params: TxParams
            The transaction parameters.

        Returns
        -------
        TxParams
            The transaction details.
        """
        return {
            **self._get_transaction_details(
                tx_function, tx_params, from_address, to_address
            ),
            "data": tx_function.buildTransaction(tx_params)["data"],
        }

    # submit transaction
    def _submit_transaction(
        self,
        tx_function: ContractFunction,
        tx_params: TxParams,
        from_address: str,
        to_address: str,
        key: str,
    ) -> str:
        """
        Submits the transaction for the transaction.

        Parameters
        ----------
        tx_function: ContractFunction
            The transaction function.
        tx_params: TxParams
            The transaction parameters.
        from_address: str
            The from address.
        to_address: str
            The to address.
        key: str
            The private key.

        Returns
        -------
        str
            The transaction hash.
        """
        tx_details = self._build_transaction(
            tx_function, tx_params, from_address, to_address
        )
        return self._get_transaction_hash(tx_details, key)



"""
translating Uniswap v3 contract values

(c) Copyright Bprotocol foundation 2023. 
Licensed under MIT

NOTE: this class is not part of the API of the Carbon protocol, and you must expect breaking
changes even in minor version updates. Use at your own risk.
"""
__VERSION__ = "1.1"
__DATE__ = "19/Apr/2023"

from math import sqrt
from dataclasses import dataclass, InitVar, asdict


@dataclass(frozen=True)
class Univ3Calculator:
    __VERSION__ = __VERSION__
    __DATE__ = __DATE__

    FEE100 = 100
    FEE500 = 500
    FEE3000 = 3000
    FEE10000 = 10000

    TICKSZ = {FEE100: 1, FEE500: 10, FEE3000: 60, FEE10000: 200}

    Q96 = 2**96
    Q192 = 2**192

    tkn0: str
    tkn1: str
    sp96: int  # sqrt_price_q96
    tick: int
    liquidity: int
    fee_const: int

    tkn0decv: InitVar[int] = None
    tkn1decv: InitVar[int] = None

    @classmethod
    def from_dict(cls, d, fee_const):
        return cls(
            tkn0=d["token0"],
            tkn1=d["token1"],
            sp96=d["sqrt_price_q96"],
            tick=d["tick"],
            liquidity=d["liquidity"],
            fee_const=fee_const,
        )

    ADDRDEC = dict(
        USDC=("0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48", 6),
        WETH=("0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2", 18),
    )
    ADDRR = {v[0]: k for k, v in ADDRDEC.items()}

    def __post_init__(self, tkn0decv=None, tkn1decv=None):
        try:
            super().__setattr__("tkn0", self.ADDRR[self.tkn0])
            super().__setattr__("tkn1", self.ADDRR[self.tkn1])
        except Exception as e:
            pass

        if tkn0decv is None:
            super().__setattr__("_tkn0dec", self.ADDRDEC[self.tkn0][1])
        else:
            super().__setattr__("_tkn0dec", tkn0decv)
        if tkn1decv is None:
            super().__setattr__("_tkn1dec", self.ADDRDEC[self.tkn1][1])
        else:
            super().__setattr__("_tkn1dec", tkn1decv)
        super().__setattr__("sp96", int(self.sp96))
        super().__setattr__("tick", int(self.tick))
        super().__setattr__("liquidity", int(self.liquidity))
        super().__setattr__("fee_const", int(self.fee_const))
        assert self.fee_const in {
            self.FEE100,
            self.FEE500,
            self.FEE3000,
            self.FEE10000,
        }, "fee not one of the FEEXXX constants {self.fee_const}"
        assert not self.tkn0dec is None, "tkn0dec is None"
        assert not self.tkn1dec is None, "tkn1dec is None"

    @property
    def pair(self):
        """the directed pair tknb/tknq = tkn0/tkn1"""
        return f"{self.tkn0}/{self.tkn1}"

    @property
    def fee(self):
        """fee in basis points"""
        return self.fee_const / 1000000

    @property
    def ticksize(self):
        """tick size"""
        return self.TICKSZ[self.fee_const]

    @property
    def tickab(self):
        """returns the tick values of Pa and Pb"""
        ticka = (self.tick // self.ticksize) * self.ticksize
        return ticka, ticka + self.ticksize

    @property
    def tkn0dec(self):
        """token 0 decimals"""
        return self._tkn0dec

    @property
    def tkn1dec(self):
        """token 1 decimals"""
        return self._tkn1dec

    @property
    def papb_raw(self):
        """raw Pa and Pb values (1.0001**tickab)"""
        ta, tb = self.tickab
        return (1.0001**ta, 1.0001**tb)

    @property
    def papb_tkn1_per_tkn0(self):
        """Pa and Pb values in units of token 1 per token 0"""
        par, pbr = self.papb_raw
        return (par * self.decf, pbr * self.decf)

    papb = papb_tkn1_per_tkn0

    @property
    def papb_tkn0_per_tkn1(self):
        """Pa and Pb values in units of token 0 per token 1"""
        pa, pb = self.papb_tkn1_per_tkn0
        return (1 / pa, 1 / pb)

    @property
    def dec_factor_wei0_per_wei1(self):
        """token wei of token 0 per token wei of token 1 at price=1"""
        return 10 ** (self.tkn0dec - self.tkn1dec)

    decf = dec_factor_wei0_per_wei1

    @classmethod
    def _price_f(cls, sp96):
        """price tkn1 per tkn0 in wei units"""
        return sp96**2 / cls.Q192

    @property
    def price_tkn1_per_tkn0(self):
        """price of token 1 per token 0 in token units"""
        return self._price_f(self.sp96) * self.decf

    p = price_tkn1_per_tkn0

    @property
    def price_tkn0_per_tkn1(self):
        """price of token 0 per token 1 in token units"""
        return 1 / self.price_tkn1_per_tkn0

    @property
    def price_convention(self):
        return f"{self.tkn0}/{self.tkn1} [{self.tkn1} per {self.tkn0}]"

    @classmethod
    def sp96_from_price(cls, price):
        """calculate sp96 = sqrt(p) * Q96"""
        return sqrt(price) * cls.Q96

    @property
    def L(self):
        """the Uniswap L value, in token units; L**2=k, and k=xy where x,y are virtual token amounts"""
        return self.liquidity / 10 ** (0.5 * (self.tkn0dec + self.tkn1dec))

    @property
    def Lsquared(self):
        """convenience methods for self.L**2 in token units"""
        return self.L**2

    L2 = Lsquared

    @property
    def k(self):
        """x*y=k, where x,y are virtual token amounts; k=L**2"""
        return self.L**2

    @property
    def kbar(self):
        """kbar = sqrt(k) = L; kbar = sqrt(xy)"""
        return self.L

    def tkn0reserve(self, *, incltoken=False):
        """returns the token 0 reserve in token units"""
        _, pb = self.papb
        sqrtPb = sqrt(pb)
        sqrtP = sqrt(self.p)
        tkn0reserve = self.liquidity * (sqrtPb - sqrtP) / (sqrtP * sqrtPb)
        tkn0reserve *= self.decf
        if not incltoken:
            return tkn0reserve
        return (tkn0reserve, self.tkn0)

    def tkn1reserve(self, *, incltoken=False):
        """returns the token 1 reserve in token units"""
        pa, _ = self.papb
        sqrtPa = sqrt(pa)
        sqrtP = sqrt(self.p)
        tkn1reserve = self.liquidity * (sqrtP - sqrtPa)
        tkn1reserve *= self.decf
        if not incltoken:
            return tkn1reserve
        return (tkn1reserve, self.tkn1)

    def tvl(self, *, astkn0=False, incltoken=False):
        """returns the total value locked in the pool"""
        tkn0reserve, tkn1reserve = self.tkn0reserve(), self.tkn1reserve()
        tvl_tkn0 = tkn0reserve + tkn1reserve * self.price_tkn0_per_tkn1
        tvl = tvl_tkn0 if astkn0 else tvl_tkn0 * self.price_tkn1_per_tkn0
        if not incltoken:
            return tvl
        return (tvl, self.tkn0 if astkn0 else self.tkn1)

    def cpc_params(self, **kwargs):
        """
        returns a kwarg dict suitable for CPC.from_univ3

        :kwargs:        additional kwargs to return
        """
        pa, pb = self.papb
        result = dict(
            Pmarg=self.p,
            uniL=self.L,
            uniPa=pa,
            uniPb=pb,
            pair=self.pair,
            fee=self.fee,
            **kwargs,
        )
        return result

    def info(self):
        pa, pb = self.papb
        p = self.p
        s = f"Uniswap v3 Range {self.tkn0}/{self.tkn1} (fee={self.fee*100:.02f}%)\n"
        s += f"  Pa = {pa:12,.3f}   P={p:12,.3f}   Pb = {pb:12,.3f} {self.tkn1} per {self.tkn0}\n"
        s += f"1/Pa = {1/pa:12,.3f} 1/P={1/p:12,.3f} 1/Pb = {1/pb:12,.3f} {self.tkn0} per {self.tkn1}\n---\n"

        s += f" full P = {p}, full 1/P = {1/p}\n"
        return s


U3 = Univ3Calculator



"""
translating Uniswap v3 contract values

(c) Copyright Bprotocol foundation 2023. 
Licensed under MIT

NOTE: this class is not part of the API of the Carbon protocol, and you must expect breaking
changes even in minor version updates. Use at your own risk.
"""
__VERSION__ = "1.4" 
__DATE__ = "07/May/2023"

from math import sqrt
from dataclasses import dataclass, InitVar, asdict


@dataclass(frozen=True)
class Univ3Calculator():
    __VERSION__ = __VERSION__
    __DATE__ = __DATE__

    FEE100      = 100
    FEE500      = 500
    FEE3000     = 3000
    FEE10000    = 10000

    TICKSZ = {FEE100: 1, FEE500: 10, FEE3000: 60, FEE10000: 200}

    Q96 = 2**96
    Q192 = 2**192
    
    tkn0: str
    tkn1: str
    sp96: int # sqrt_price_q96
    tick: int
    liquidity: int
    fee_const: int 

    tkn0decv: InitVar[int] = None
    tkn1decv: InitVar[int] = None
    addrdec: InitVar[dict] = None
    ADDRDEC = dict(
        # only for testing
        USDC = ("0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48", 6),
        WETH = ("0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2", 18),
    )

    @classmethod
    def from_dict(cls, d, fee_const, *, addrdec=None, tkn0decv=None, tkn1decv=None):
        """
        alternative constructor from a dictionary
        
        :d:             dict with keys: token0 [address], token1 [address], sqrt_price_q96, tick, liquidity
        :fee_const:     fee constant (FEE100, ...)
        :tkn0decv:      optional token0 decimals value (eg 6, 18)
        :tkn1decv:      optional token1 decimals value (eg 6, 18)
        :addrdec:       optional dictionary of token address to decimals (eg {"0x123...": 18})
        """
        return cls(
            tkn0 = d["token0"],
            tkn1 = d["token1"],
            sp96 = d["sqrt_price_q96"],
            tick = d["tick"],
            liquidity = d["liquidity"],
            fee_const = fee_const,
            addrdec = addrdec,
            tkn0decv = tkn0decv,
            tkn1decv = tkn1decv,
        )


    ADDRR = {v[0]:k for k,v in ADDRDEC.items()}
    
    class DecimalsMissingError(Exception): pass
    
    def __post_init__(self, tkn0decv=None, tkn1decv=None, addrdec=None):
        
        #print("[Univ3Calculator] post_init", tkn0decv, tkn0decv)

        if addrdec is not None:
            self.ADDRDEC.update(addrdec)
            self.ADDRR.update({v[0]:k for k,v in addrdec.items()})

        try:
            super().__setattr__('tkn0', self.ADDRR[self.tkn0])
            super().__setattr__('tkn1', self.ADDRR[self.tkn1])
        except Exception as e:
            pass

        if tkn0decv is None:
            try:
                super().__setattr__('_tkn0dec', self.ADDRDEC[self.tkn0][1])
            except KeyError:
                raise self.DecimalsMissingError(f"must provide tkn0decv for {self.tkn0}")
        else:
            super().__setattr__('_tkn0dec', tkn0decv)
        
        if tkn1decv is None:
            try:
                super().__setattr__('_tkn1dec', self.ADDRDEC[self.tkn1][1])
            except KeyError:
                raise self.DecimalsMissingError(f"must provide tkn1decv for {self.tkn1}")
        else:
            super().__setattr__('_tkn1dec', tkn1decv)
        
        super().__setattr__('sp96', int(self.sp96)) 
        super().__setattr__('tick', int(self.tick)) 
        super().__setattr__('liquidity', int(self.liquidity)) 
        super().__setattr__('fee_const', int(self.fee_const)) 
        assert self.fee_const in {self.FEE100, self.FEE500, self.FEE3000, self.FEE10000}, "fee not one of the FEEXXX constants {self.fee_const}"
        assert not self.tkn0dec is None, "tkn0dec is None"
        assert not self.tkn1dec is None, "tkn1dec is None"

    @property
    def pair(self):
        """the directed pair tknb/tknq = tkn0/tkn1"""
        return f"{self.tkn0}/{self.tkn1}"
    
    @property
    def fee(self):
        """fee in basis points"""
        return self.fee_const/1000000
    
    @property
    def ticksize(self):
        """tick size"""
        return self.TICKSZ[self.fee_const]
    
    @property
    def tickab(self):
        """returns the tick values of Pa and Pb"""
        ticka = (self.tick // self.ticksize) * self.ticksize
        return ticka, ticka + self.ticksize
    
    @property
    def tkn0dec(self):
        """token 0 decimals"""
        return self._tkn0dec
    
    @property
    def tkn1dec(self):
        """token 1 decimals"""
        return self._tkn1dec
    
    @property
    def papb_raw(self):
        """raw Pa and Pb values (1.0001**tickab)"""
        ta, tb = self.tickab
        return (1.0001**ta, 1.0001**tb)
    
    @property
    def papb_tkn1_per_tkn0(self):
        """Pa and Pb values in units of token 1 per token 0"""
        par,pbr = self.papb_raw
        return (par*self.decf, pbr*self.decf) 
    papb = papb_tkn1_per_tkn0

    @property
    def papb_tkn0_per_tkn1(self):
        """Pa and Pb values in units of token 0 per token 1"""
        pa,pb = self.papb_tkn1_per_tkn0
        return (1/pa, 1/pb)
    
    @property
    def dec_factor_wei0_per_wei1(self):
        """token wei of token 0 per token wei of token 1 at price=1"""
        return 10**(self.tkn0dec-self.tkn1dec)
    decf = dec_factor_wei0_per_wei1

    @classmethod
    def _price_f(cls, sp96):
        """price tkn1 per tkn0 in wei units"""
        return sp96 ** 2 / cls.Q192
    
    @property
    def price_tkn1_per_tkn0(self):
        """price of token 1 per token 0 in token units"""
        return self._price_f(self.sp96) * self.decf
    p = price_tkn1_per_tkn0
    
    @property
    def price_tkn0_per_tkn1(self):
        """price of token 0 per token 1 in token units"""
        return 1/self.price_tkn1_per_tkn0

    @property
    def price_convention(self):
        return f"{self.tkn0}/{self.tkn1} [{self.tkn1} per {self.tkn0}]"
    
    @classmethod
    def sp96_from_price(cls, price):
        """calculate sp96 = sqrt(p) * Q96"""
        return sqrt(price) * cls.Q96
    
    @property
    def L(self):
        """the Uniswap L value, in token units; L**2=k, and k=xy where x,y are virtual token amounts"""
        return self.liquidity/10**(0.5*(self.tkn0dec+self.tkn1dec)) if self.liquidity!=0 else 0
    
    @property
    def Lsquared(self):
        """convenience methods for self.L**2 in token units"""
        return self.L**2
    L2 = Lsquared
    
    @property
    def k(self):
        """x*y=k, where x,y are virtual token amounts; k=L**2"""
        return self.L**2
    
    @property
    def kbar(self):
        """kbar = sqrt(k) = L; kbar = sqrt(xy)"""
        return self.L
    
    def tkn0reserve(self, *, incltoken=False):
        """returns the token 0 reserve in token units"""
        _, pb = self.papb
        sqrtPb = sqrt(pb)
        sqrtP = sqrt(self.p)
        tkn0reserve = self.liquidity * (sqrtPb - sqrtP) / (sqrtP * sqrtPb) if (sqrtP * sqrtPb)!=0 else 0
        tkn0reserve *= self.decf
        if not incltoken:
            return tkn0reserve
        return (tkn0reserve, self.tkn0)
    
    def tkn1reserve(self, *, incltoken=False):
        """returns the token 1 reserve in token units"""
        pa, _ = self.papb
        sqrtPa = sqrt(pa)
        sqrtP = sqrt(self.p)
        tkn1reserve = self.liquidity * (sqrtP - sqrtPa)
        tkn1reserve *= self.decf
        if not incltoken:
            return tkn1reserve
        return (tkn1reserve, self.tkn1)
    
    def tvl(self, *, astkn0=False, incltoken=False):
        """returns the total value locked in the pool"""
        tkn0reserve, tkn1reserve = self.tkn0reserve(), self.tkn1reserve()
        tvl_tkn0 = tkn0reserve + tkn1reserve*self.price_tkn0_per_tkn1
        tvl = tvl_tkn0 if astkn0 else tvl_tkn0*self.price_tkn1_per_tkn0
        if not incltoken:
            return tvl
        return (tvl, self.tkn0 if astkn0 else self.tkn1)
    
    def cpc_params(self, **kwargs):
        """
        returns a kwarg dict suitable for CPC.from_univ3
        
        :kwargs:        additional kwargs to return
        """
        pa,pb = self.papb
        pm = self.p
        pmar = pm/pa - 1
        pmbr = pm/pb - 1
        #print("[cpc_params]", pa, pm, pb, pmar, pmbr)
        if pmar < 0:
            #print("[cpc_params] pmar<0; asserting just rounding", pa, pm, pmar)
            assert pmar > -1e-10, f"pm below pa beyond rounding error [{pm}, {pa}, {pmar}]"
        if abs(pmar)<1e-10:
            #print("[cpc_params] setting pm to pa", pa, pm, pmar)
            pm = pa
        
        if pmbr < 0:
            #print("[cpc_params] pmbr>0; asserting just rounding", pm, pb, pmbr)
            assert pmbr < 1e-10, f"pm abve pb beyond rounding error [{pm}, {pb}, {pmbr}]"
        if abs(pmbr)<1e-10:
            #print("[cpc_params] setting pm to pb", pm, pb, pmbr)
            pm = pb
            
        result = dict(
            Pmarg = pm,
            uniL = self.L,
            uniPa = pa,
            uniPb = pb,
            pair = self.pair,
            fee = self.fee,
            **kwargs,
        )
        return result
    
    def info(self):
        pa, pb = self.papb
        p = self.p
        s  = f"Uniswap v3 Range {self.tkn0}/{self.tkn1} (fee={self.fee*100:.02f}%)\n"
        s += f"  Pa = {pa:12,.3f}   P={p:12,.3f}   Pb = {pb:12,.3f} {self.tkn1} per {self.tkn0}\n"
        s += f"1/Pa = {1/pa:12,.3f} 1/P={1/p:12,.3f} 1/Pb = {1/pb:12,.3f} {self.tkn0} per {self.tkn1}\n---\n"
    
        s += f" full P = {p}, full 1/P = {1/p}\n"
        return s

U3 = Univ3Calculator


# coding=utf-8
"""
Contains the utils functions for events

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""
from typing import Any, Union, Dict, Set
from typing import List

from hexbytes import HexBytes
from web3.datastructures import AttributeDict

from fastlane_bot.events.managers.manager import Manager


def filter_latest_events(
    mgr: Manager, events: List[List[AttributeDict]]
) -> List[AttributeDict]:
    """
    This function filters out the latest events for each pool. Given a nested list of events, it iterates through all events
    and keeps track of the latest event (i.e., with the highest block number) for each pool. The key used to identify each pool
    is derived from the event data using manager's methods.

    Args:
        mgr (Base): A Base object that provides methods to handle events and their related pools.
        events (List[List[AttributeDict]]): A nested list of events, where each event is an AttributeDict that includes
        the event data and associated metadata.

    Returns:
        List[AttributeDict]: A list of events, each representing the latest event for its corresponding pool.
    """
    latest_entry_per_pool = {}
    all_events = [event for event_list in events for event in event_list]

    # Handles the case where multiple pools are created in the same block
    all_events.reverse()

    for event in all_events:
        pool_type = mgr.pool_type_from_exchange_name(
            mgr.exchange_name_from_event(event)
        )
        if pool_type:
            key = pool_type.unique_key()
        else:
            continue
        if key == "cid":
            key = "id"
        elif key == "tkn1_address":
            if event["args"]["pool"] != mgr.cfg.BNT_ADDRESS:
                key = "pool"
            else:
                key = "tkn_address"
        unique_key = event[key] if key in event else event["args"][key]

        if unique_key in latest_entry_per_pool:
            if event["blockNumber"] > latest_entry_per_pool[unique_key]["blockNumber"]:
                latest_entry_per_pool[unique_key] = event
            elif (
                event["blockNumber"] == latest_entry_per_pool[unique_key]["blockNumber"]
            ):
                if (
                    event["transactionIndex"]
                    == latest_entry_per_pool[unique_key]["transactionIndex"]
                ):
                    if (
                        event["logIndex"]
                        > latest_entry_per_pool[unique_key]["logIndex"]
                    ):
                        latest_entry_per_pool[unique_key] = event
                elif (
                    event["transactionIndex"]
                    > latest_entry_per_pool[unique_key]["transactionIndex"]
                ):
                    latest_entry_per_pool[unique_key] = event
                else:
                    continue
            else:
                continue
        else:
            latest_entry_per_pool[unique_key] = event

    return list(latest_entry_per_pool.values())


def complex_handler(obj: Any) -> Union[Dict, str, List, Set, Any]:
    """
    This function aims to handle complex data types, such as web3.py's AttributeDict, HexBytes, and native Python collections
    like dict, list, tuple, and set. It recursively traverses these collections and converts their elements into more "primitive"
    types, making it easier to work with these elements or serialize the data into JSON.

    Args:
        obj (Any): The object to be processed. This can be of any data type, but the function specifically handles AttributeDict,
        HexBytes, dict, list, tuple, and set.

    Returns:
        Union[Dict, str, List, Set, Any]: Returns a "simplified" version of the input object, where AttributeDict is converted
        into dict, HexBytes into str, and set into list. For dict, list, and tuple, it recursively processes their elements.
        If the input object does not match any of the specified types, it is returned as is.
    """
    if isinstance(obj, AttributeDict):
        return dict(obj)
    elif isinstance(obj, HexBytes):
        return obj.hex()
    elif isinstance(obj, bytes):
        return obj.hex()
    elif isinstance(obj, dict):
        return {k: complex_handler(v) for k, v in obj.items()}
    elif isinstance(obj, (list, tuple)):
        return [complex_handler(i) for i in obj]
    elif isinstance(obj, set):
        return list(obj)
    else:
        return obj



# coding=utf-8
"""
Contains the factory class for pools. This class is responsible for creating pools.

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""
from fastlane_bot.events.pools.base import Pool


class PoolFactory:
    """
    Factory class for creating pools.
    """

    def __init__(self):
        self._creators = {}

    def register_format(self, format_name: str, creator: Pool) -> None:
        """
        Register a pool type.

        Parameters
        ----------
        format_name : str
            The name of the pool type.
        creator : Pool
            The pool class.
        """
        self._creators[format_name] = creator

    def get_pool(self, format_name: str) -> Pool:
        """
        Get a pool.

        Parameters
        ----------
        format_name : str
            The name of the pool type.
        """
        creator = self._creators.get(format_name)
        if not creator:
            if format_name:
                raise ValueError(format_name)
            else:
                # raise ValueError("format_name")
                pass
        return creator


# create an instance of the factory
pool_factory = PoolFactory()



# coding=utf-8
"""
Contains the pools classes

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""
import pandas as pd

from fastlane_bot.events.pools.bancor_v3 import BancorV3Pool
from fastlane_bot.events.pools.carbon_v1 import CarbonV1Pool
from fastlane_bot.events.pools.factory import pool_factory
from fastlane_bot.events.pools.sushiswap_v2 import SushiswapV2Pool
from fastlane_bot.events.pools.uniswap_v2 import UniswapV2Pool
from fastlane_bot.events.pools.uniswap_v3 import UniswapV3Pool

# register your pool types
pool_factory.register_format("uniswap_v3", UniswapV3Pool)
pool_factory.register_format("uniswap_v2", UniswapV2Pool)
pool_factory.register_format("sushiswap_v2", SushiswapV2Pool)
pool_factory.register_format("bancor_v3", BancorV3Pool)
pool_factory.register_format("carbon_v1", CarbonV1Pool)



# coding=utf-8
"""
Bancor V3 triangular arbitrage finder mode

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""
import math
from typing import Union, List, Tuple, Any, Iterable

from fastlane_bot.modes.base_triangle import ArbitrageFinderTriangleBase
from fastlane_bot.tools.cpc import CPCContainer, T, ConstantProductCurve
from fastlane_bot.tools.optimizer import CPCArbOptimizer


class ArbitrageFinderTriangleSingleBancor3(ArbitrageFinderTriangleBase):
    """
    Bancor V3 triangular arbitrage finder mode
    """

    arb_mode = "single_triangle_bancor3"

    def find_arbitrage(self, candidates: List[Any] = None, ops: Tuple = None, best_profit: float = 0, profit_src: float = 0) -> Union[List, Tuple]:
        """
        see base.py
        """
        if self.base_exchange != "bancor_v3":
            self.ConfigObj.logger.warning(
                f"base_exchange must be bancor_v3 for {self.arb_mode}, setting it to bancor_v3"
            )
            self.base_exchange = "bancor_v3"

        self.ConfigObj.logger.info(
            f"flashloan_tokens for arb_mode={self.arb_mode} will be overwritten. "
        )
        self.flashloan_tokens = self.CCm.byparams(exchange="bancor_v3").tknys()

        if candidates is None:
            candidates = []

        # Get combinations of flashloan tokens
        combos = self.get_combos(
            self.flashloan_tokens, self.CCm, arb_mode=self.arb_mode
        )

        # Get the miniverse combinations
        all_miniverses = self.get_miniverse_combos(combos)

        if len(all_miniverses) == 0:
            return None

        # Check each source token and miniverse combination
        for src_token, miniverse in all_miniverses:
            r = None

            try:
                # Run main flow with the new set of curves
                (
                    profit_src,
                    trade_instructions,
                    trade_instructions_df,
                    trade_instructions_dic,
                ) = self.run_main_flow(miniverse, src_token)

                # Get the cids of the carbon pools
                carbon_cids = [
                    curve.cid
                    for curve in miniverse
                    if curve.params.get("exchange") == "carbon_v1"
                ]

                if carbon_cids:

                    # Get the new set of curves
                    new_curves = self.get_mono_direction_carbon_curves(
                        miniverse=miniverse, trade_instructions_df=trade_instructions_df
                    )

                    # Rerun main flow with the new set of curves
                    (
                        profit_src,
                        trade_instructions,
                        trade_instructions_df,
                        trade_instructions_dic,
                    ) = self.run_main_flow(new_curves, src_token)
            except Exception:
                continue

            # Get the candidate ids
            cids = [ti["cid"] for ti in trade_instructions_dic]

            # Calculate the profit
            profit = self.calculate_profit(src_token, profit_src, self.CCm, cids)

            if str(profit) == "nan":
                self.ConfigObj.logger.debug("profit is nan, skipping")
                continue

            # Handle candidates based on conditions
            candidates += self.handle_candidates(
                best_profit,
                profit,
                trade_instructions_df,
                trade_instructions_dic,
                src_token,
                trade_instructions,
            )

            # Find the best operations
            best_profit, ops = self.find_best_operations(
                best_profit,
                ops,
                profit,
                trade_instructions_df,
                trade_instructions_dic,
                src_token,
                trade_instructions,
            )

        return candidates if self.result == self.AO_CANDIDATES else ops

    def get_tkn(self, pool: ConstantProductCurve, tkn_num: int) -> str:
        """
        Gets the token ID from a pool object

        Parameters
        ----------
        pool: ConstantProductCurve
            The ConstantProductCurve object
        tkn_num: int
            The token number to get, either 0 or 1

        Returns
        -------
        str
        """
        return pool.pair.split("/")[tkn_num]

    @staticmethod
    def get_fee_safe(fee: int or float):
        """
        Fixes the format of the fee if the fee is in PPM format

        Parameters
        ----------
        fee: int or float

        Returns
        -------
        float
        """
        if fee > 1:
            fee = fee / 1000000
        return fee

    def get_exact_pools(self, cids: List[str]) -> List[CPCContainer]:
        """
        Gets the specific pools that will be used for calculations. It does this inefficiently to preserve the order.

        Parameters
        ----------
        cids: List

        Returns
        -------
        List
        """

        pools = [curve for curve in self.CCm if curve.cid == cids[0]]
        pools += [curve for curve in self.CCm if curve.cid == cids[1]]
        pools += [curve for curve in self.CCm if curve.cid == cids[2]]
        return pools

    def get_optimal_arb_trade_amts(self, cids: List[str], flt: str) -> float:
        """
        Gets the optimal trade 0 amount for a triangular arb cycle

        Parameters
        ----------
        cids: List
        flt: str

        Returns
        -------
        float
        """
        pools = self.get_exact_pools(cids=cids)
        tkn0 = self.get_tkn(pool=pools[0], tkn_num=0)
        tkn1 = self.get_tkn(pool=pools[0], tkn_num=1)
        tkn2 = self.get_tkn(pool=pools[1], tkn_num=0)
        tkn5 = self.get_tkn(pool=pools[2], tkn_num=0)
        p0t0 = pools[0].x_act if tkn0 == flt else pools[0].y_act
        p0t1 = pools[0].y_act if tkn0 == flt else pools[0].x_act
        p2t1 = pools[2].x_act if tkn5 == flt else pools[2].y_act
        p2t0 = pools[2].y_act if tkn5 == flt else pools[2].x_act
        fee1 = self.get_fee_safe(pools[1].fee)

        if pools[1].params.exchange == "carbon_v1":
            return self.get_exact_input_with_carbon(p0t0, p0t1, p2t0, p2t1, pools[1])

        p1t0 = pools[1].x if tkn1 == tkn2 else pools[1].y
        p1t1 = pools[1].y if tkn1 == tkn2 else pools[1].x
        fee0 = 0
        fee2 = 0
        return self.max_arb_trade_in_constant_product(p0t0, p0t1, p1t0, p1t1, p2t0, p2t1, fee0=fee0, fee1=fee1, fee2=fee2)

    def get_exact_input_with_carbon(self, p0t0: float, p0t1: float, p2t0: float, p2t1: float, carbon_pool: ConstantProductCurve) -> float:
        """
        Gets the optimal trade 0 amount for a triangular arb cycle with a single Carbon order in the middle

        Parameters
        ----------
        p0t0: float
        p0t1: float
        p2t0: float
        p2t1: float
        carbon_pool: ConstantProductCurve

        Returns
        -------
        float
        """
        y = carbon_pool.params['y']
        z = carbon_pool.params['yint']
        A = carbon_pool.params['A']
        B = carbon_pool.params['B']
        C = (B * z + A * y) ** 2
        D = B * A * z + A ** 2 * y
        return self.max_arb_trade_in_cp_carbon_cp(p0t0, p0t1, p2t0, p2t1, C, D, z)

    @staticmethod
    def max_arb_trade_in_cp_carbon_cp(p0t0: float, p0t1: float, p2t0: float, p2t1: float, C: float, D: float, z: float) -> float:
        """
        Equation to solve optimal trade input for a constant product -> Carbon order -> constant product route.
        Parameters
        ----------
        p0t0: float
        p0t1: float
        p2t0: float
        p2t1: float
        C: float
        D: float
        z: float
        Returns
        -------
        float

        """
        trade_input = (z * (-p0t0 * p2t0 * z + math.sqrt(C * p0t0 * p2t0 * p0t1 * p2t1))) / (p0t1 * C + p0t1 * D * p2t0 + z ** 2 * p2t0)
        return trade_input

    @staticmethod
    def max_arb_trade_in_constant_product(p0t0, p0t1, p1t0, p1t1, p2t0, p2t1, fee0, fee1, fee2):
        """
        Equation to solve optimal trade input for a constant product -> constant product -> constant product route.
        Parameters
        ----------
        p0t0: float
        p0t1: float
        p1t0: float
        p1t1: float
        p2t0: float
        p2t1: float
        fee0: float
        fee1: float
        fee2: float
        Returns
        -------
        float

        """
        val = (-p1t0*p2t0*p0t0 + (p1t0*p2t0*p0t0*p1t1*p2t1*p0t1*(-fee1*fee2*fee0 + fee1*fee2 + fee1*fee0 - fee1 + fee2*fee0 - fee2 - fee0 + 1)) ** 0.5)/(p1t0*p2t0 - p2t0*p0t1*fee0 + p2t0*p0t1 + p1t1*p0t1*fee1*fee0 - p1t1*p0t1*fee1 - p1t1*p0t1*fee0 + p1t1*p0t1)
        return val

    @staticmethod
    def run_main_flow(
        miniverse: List, src_token: str
    ) -> Tuple[float, Any, Any, Any]:
        """
        Run the main flow of the arbitrage finder.

        Parameters
        ----------
        miniverse : list
            List of curves.
        src_token : str
            Source token.

        Returns
        -------
        tuple
            Tuple of profit, trade instructions, trade instructions dataframe and trade instructions dictionary.

        """

        # Instantiate the container and optimizer objects
        CC_cc = CPCContainer(miniverse)
        O = CPCArbOptimizer(CC_cc)

        # Perform the optimization
        r = O.margp_optimizer(src_token)

        # Get the profit in the source token
        profit_src = -r.result

        # Get trade instructions in different formats
        trade_instructions_df = r.trade_instructions(O.TIF_DFAGGR)
        trade_instructions_dic = r.trade_instructions(O.TIF_DICTS)
        trade_instructions = r.trade_instructions()

        return (
            profit_src,
            trade_instructions,
            trade_instructions_df,
            trade_instructions_dic,
        )

    def get_miniverse_combos(self, combos: Iterable) -> List[Tuple[str, List]]:
        """
        Get the miniverse combinations for a list of token pairs.

        Parameters
        ----------
        combos : list
            List of token pairs.

        Returns
        -------
        list
            List of miniverse combinations.

        """
        all_miniverses = []
        for tkn0, tkn1 in combos:
            external_curves = self.CCm.bypairs(f"{tkn0}/{tkn1}")
            external_curves += self.CCm.bypairs(f"{tkn1}/{tkn0}")
            external_curves = list(set(external_curves))
            carbon_curves = [
                curve
                for curve in external_curves
                if curve.params.get("exchange") == "carbon_v1"
            ]
            external_curves = [
                curve
                for curve in external_curves
                if curve.params.get("exchange") != "carbon_v1"
            ]
            if not external_curves and not carbon_curves:
                continue

            bancor_v3_curve_0 = (
                self.CCm.bypairs(f"{T.BNT}/{tkn0}")
                .byparams(exchange="bancor_v3")
                .curves
            )
            bancor_v3_curve_1 = (
                self.CCm.bypairs(f"BNT-FF1C/{tkn1}")
                .byparams(exchange="bancor_v3")
                .curves
            )
            if bancor_v3_curve_0 is None or bancor_v3_curve_1 is None:
                continue
            if len(bancor_v3_curve_0) == 0 or len(bancor_v3_curve_1) == 0:
                continue

            miniverses = []
            if external_curves:
                for curve in external_curves:
                    miniverses += [bancor_v3_curve_0 + bancor_v3_curve_1 + [curve]]
            if carbon_curves:
                miniverses += [bancor_v3_curve_0 + bancor_v3_curve_1 + carbon_curves]

            if len(miniverses) > 0:
                all_miniverses += list(zip([T.BNT] * len(miniverses), miniverses))
        return all_miniverses
